@@ -43,6 +43,7 @@
 #include <private/qsgadaptationlayer_p.h>
 #include <private/qquickitem_p.h>
 #include <private/qguiapplication_p.h>
+#include <private/qquickmultipointhandler_p_p.h>
 #include <private/qquickwindow_p.h>
 #include <QEvent>
 #include <QMouseEvent>
@@ -84,12 +85,21 @@ Q_LOGGING_CATEGORY(lcPinchHandler, "qt.quick.handler.pinch")
 
     \image touchpoints-pinchhandler.png
 
-    \sa PinchArea
+    \note The pinch begins when the number of fingers pressed is between
+    \l {MultiPointHandler::minimumPointCount}{minimumPointCount} and
+    \l {MultiPointHandler::maximumPointCount}{maximumPointCount}, inclusive.
+    Until then, PinchHandler tracks the positions of any pressed fingers,
+    but if it's a disallowed number, it does not scale or rotate
+    its \l target, and the \l active property remains \c false.
+
+    \sa PinchArea, QPointerEvent::pointCount()
 */
 
 QQuickPinchHandler::QQuickPinchHandler(QQuickItem *parent)
     : QQuickMultiPointHandler(parent, 2)
 {
+    // Tell QQuickPointerDeviceHandler::wantsPointerEvent() to ignore button state
+    d_func()->acceptedButtons = Qt::NoButton;
 }
 
 /*!
@@ -152,58 +162,16 @@ void QQuickPinchHandler::setMaximumRotation(qreal maximumRotation)
     emit maximumRotationChanged();
 }
 
-#if QT_DEPRECATED_SINCE(5, 12)
-void QQuickPinchHandler::warnAboutMinMaxDeprecated() const
-{
-    qmlWarning(this) << "min and max constraints are now part of the xAxis and yAxis properties";
-}
-
-void QQuickPinchHandler::setMinimumX(qreal minX)
-{
-    warnAboutMinMaxDeprecated();
-    if (qFuzzyCompare(m_minimumX, minX))
-        return;
-    m_minimumX = minX;
-    emit minimumXChanged();
-}
-
-void QQuickPinchHandler::setMaximumX(qreal maxX)
-{
-    warnAboutMinMaxDeprecated();
-    if (qFuzzyCompare(m_maximumX, maxX))
-        return;
-    m_maximumX = maxX;
-    emit maximumXChanged();
-}
-
-void QQuickPinchHandler::setMinimumY(qreal minY)
-{
-    warnAboutMinMaxDeprecated();
-    if (qFuzzyCompare(m_minimumY, minY))
-        return;
-    m_minimumY = minY;
-    emit minimumYChanged();
-}
-
-void QQuickPinchHandler::setMaximumY(qreal maxY)
-{
-    warnAboutMinMaxDeprecated();
-    if (qFuzzyCompare(m_maximumY, maxY))
-        return;
-    m_maximumY = maxY;
-    emit maximumYChanged();
-}
-#endif
-
-bool QQuickPinchHandler::wantsPointerEvent(QQuickPointerEvent *event)
+bool QQuickPinchHandler::wantsPointerEvent(QPointerEvent *event)
 {
     if (!QQuickMultiPointHandler::wantsPointerEvent(event))
         return false;
 
 #if QT_CONFIG(gestures)
-    if (const auto gesture = event->asPointerNativeGestureEvent()) {
+    if (event->type() == QEvent::NativeGesture) {
+        const auto gesture = static_cast<const QNativeGestureEvent *>(event);
         if (minimumPointCount() == 2) {
-            switch (gesture->type()) {
+            switch (gesture->gestureType()) {
             case Qt::BeginNativeGesture:
             case Qt::EndNativeGesture:
             case Qt::ZoomNativeGesture:
@@ -247,68 +215,56 @@ bool QQuickPinchHandler::wantsPointerEvent(QQuickPointerEvent *event)
     If \c enabled is true, vertical dragging is allowed.
  */
 
-/*!
-    \qmlproperty int QtQuick::PinchHandler::minimumTouchPoints
-
-    The pinch begins when this number of fingers are pressed.
-    Until then, PinchHandler tracks the positions of any pressed fingers,
-    but if it's an insufficient number, it does not scale or rotate
-    its \l target, and the \l active property will remain false.
-*/
-
 /*!
     \qmlproperty bool QtQuick::PinchHandler::active
 
-    This property is true when all the constraints (epecially \l minimumTouchPoints)
-    are satisfied and the \l target, if any, is being manipulated.
+    This property is \c true when all the constraints (epecially
+    \l {MultiPointHandler::minimumPointCount}{minimumPointCount} and
+    \l {MultiPointHandler::maximumPointCount}{maximumPointCount}) are satisfied
+    and the \l target, if any, is being manipulated.
 */
 
 void QQuickPinchHandler::onActiveChanged()
 {
     QQuickMultiPointHandler::onActiveChanged();
     if (active()) {
-        m_startMatrix = QMatrix4x4();
-        m_startAngles = angles(m_centroid.sceneGrabPosition());
-        m_startDistance = averageTouchPointDistance(m_centroid.sceneGrabPosition());
+        m_startAngles = angles(centroid().sceneGrabPosition());
+        m_startDistance = averageTouchPointDistance(centroid().sceneGrabPosition());
         m_activeRotation = 0;
         m_activeTranslation = QVector2D();
         if (const QQuickItem *t = target()) {
             m_startScale = t->scale(); // TODO incompatible with independent x/y scaling
             m_startRotation = t->rotation();
-            QVector3D xformOrigin(t->transformOriginPoint());
-            m_startMatrix.translate(float(t->x()), float(t->y()));
-            m_startMatrix.translate(xformOrigin);
-            m_startMatrix.scale(float(m_startScale));
-            m_startMatrix.rotate(float(m_startRotation), 0, 0, -1);
-            m_startMatrix.translate(-xformOrigin);
+            m_startPos = t->position();
         } else {
-            m_startScale = m_accumulatedScale;
+            m_startScale = 1;
             m_startRotation = 0;
         }
-        qCInfo(lcPinchHandler) << "activated with starting scale" << m_startScale << "rotation" << m_startRotation;
+        qCDebug(lcPinchHandler) << "activated with starting scale" << m_startScale << "rotation" << m_startRotation;
     } else {
-        qCInfo(lcPinchHandler) << "deactivated with scale" << m_activeScale << "rotation" << m_activeRotation;
+        qCDebug(lcPinchHandler) << "deactivated with scale" << m_activeScale << "rotation" << m_activeRotation;
     }
 }
 
-void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
+void QQuickPinchHandler::handlePointerEventImpl(QPointerEvent *event)
 {
     if (Q_UNLIKELY(lcPinchHandler().isDebugEnabled())) {
-        for (const QQuickHandlerPoint &p : m_currentPoints)
-            qCDebug(lcPinchHandler) << hex << p.id() << p.sceneGrabPosition() << "->" << p.scenePosition();
+        for (const QQuickHandlerPoint &p : currentPoints())
+            qCDebug(lcPinchHandler) << Qt::hex << p.id() << p.sceneGrabPosition() << "->" << p.scenePosition();
     }
     QQuickMultiPointHandler::handlePointerEventImpl(event);
 
     qreal dist = 0;
 #if QT_CONFIG(gestures)
-    if (const auto gesture = event->asPointerNativeGestureEvent()) {
-        m_centroid.reset(event->point(0));
-        switch (gesture->type()) {
+    if (event->type() == QEvent::NativeGesture) {
+        const auto gesture = static_cast<const QNativeGestureEvent *>(event);
+        mutableCentroid().reset(event, event->point(0));
+        switch (gesture->gestureType()) {
         case Qt::EndNativeGesture:
             m_activeScale = 1;
             m_activeRotation = 0;
             m_activeTranslation = QVector2D();
-            m_centroid.reset();
+            mutableCentroid().reset();
             setActive(false);
             emit updated();
             return;
@@ -331,32 +287,32 @@ void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
     } else
 #endif // QT_CONFIG(gestures)
     {
-        const bool containsReleasedPoints = event->isReleaseEvent();
-        QVector<QQuickEventPoint *> chosenPoints;
-        for (const QQuickHandlerPoint &p : m_currentPoints) {
-            QQuickEventPoint *ep = event->pointById(p.id());
-            chosenPoints << ep;
+        const bool containsReleasedPoints = event->isEndEvent();
+        QVector<QEventPoint> chosenPoints;
+        for (const QQuickHandlerPoint &p : currentPoints()) {
+            auto ep = event->pointById(p.id());
+            Q_ASSERT(ep);
+            chosenPoints << *ep;
         }
         if (!active()) {
             // Verify that at least one of the points has moved beyond threshold needed to activate the handler
             int numberOfPointsDraggedOverThreshold = 0;
             QVector2D accumulatedDrag;
-            const QVector2D currentCentroid(m_centroid.scenePosition());
-            const QVector2D pressCentroid(m_centroid.scenePressPosition());
+            const QVector2D currentCentroid(centroid().scenePosition());
+            const QVector2D pressCentroid(centroid().scenePressPosition());
 
-            QStyleHints *styleHints = QGuiApplication::styleHints();
-            const int dragThreshold = styleHints->startDragDistance();
+            const int dragThreshold = QQuickPointerHandler::dragThreshold();
             const int dragThresholdSquared = dragThreshold * dragThreshold;
 
             double accumulatedCentroidDistance = 0;     // Used to detect scale
-            if (event->isPressEvent())
+            if (event->isBeginEvent())
                 m_accumulatedStartCentroidDistance = 0;   // Used to detect scale
 
             float accumulatedMovementMagnitude = 0;
 
-            for (QQuickEventPoint *point : qAsConst(chosenPoints)) {
+            for (auto &point : chosenPoints) {
                 if (!containsReleasedPoints) {
-                    accumulatedDrag += QVector2D(point->scenePressPosition() - point->scenePosition());
+                    accumulatedDrag += QVector2D(point.scenePressPosition() - point.scenePosition());
                     /*
                        In order to detect a drag, we want to check if all points have moved more or
                        less in the same direction.
@@ -390,29 +346,30 @@ void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
                          was moved, but the relative position between each finger remained very much
                          the same). This is then used to rule out if there is a rotation or scale.
                     */
-                    QVector2D pressCentroidRelativePosition = QVector2D(point->scenePosition()) - currentCentroid;
-                    QVector2D currentCentroidRelativePosition = QVector2D(point->scenePressPosition()) - pressCentroid;
+                    QVector2D pressCentroidRelativePosition = QVector2D(point.scenePosition()) - currentCentroid;
+                    QVector2D currentCentroidRelativePosition = QVector2D(point.scenePressPosition()) - pressCentroid;
                     QVector2D centroidRelativeMovement = currentCentroidRelativePosition - pressCentroidRelativePosition;
                     accumulatedMovementMagnitude += centroidRelativeMovement.length();
 
                     accumulatedCentroidDistance += qreal(pressCentroidRelativePosition.length());
-                    if (event->isPressEvent())
-                        m_accumulatedStartCentroidDistance += qreal((QVector2D(point->scenePressPosition()) - pressCentroid).length());
+                    if (event->isBeginEvent())
+                        m_accumulatedStartCentroidDistance += qreal((QVector2D(point.scenePressPosition()) - pressCentroid).length());
                 } else {
-                    setPassiveGrab(point);
+                    setPassiveGrab(event, point);
                 }
-                if (point->state() == QQuickEventPoint::Pressed) {
-                    point->setAccepted(false); // don't stop propagation
-                    setPassiveGrab(point);
+                if (point.state() == QEventPoint::Pressed) {
+                    point.setAccepted(false); // don't stop propagation
+                    setPassiveGrab(event, point);
                 }
-                if (QQuickWindowPrivate::dragOverThreshold(point))
+                Q_D(QQuickMultiPointHandler);
+                if (d->dragOverThreshold(point))
                     ++numberOfPointsDraggedOverThreshold;
             }
 
             const bool requiredNumberOfPointsDraggedOverThreshold = numberOfPointsDraggedOverThreshold >= minimumPointCount() && numberOfPointsDraggedOverThreshold <= maximumPointCount();
-            accumulatedMovementMagnitude /= m_currentPoints.count();
+            accumulatedMovementMagnitude /= currentPoints().count();
 
-            QVector2D avgDrag = accumulatedDrag / m_currentPoints.count();
+            QVector2D avgDrag = accumulatedDrag / currentPoints().count();
             if (!xAxis()->enabled())
                 avgDrag.setX(0);
             if (!yAxis()->enabled())
@@ -424,17 +381,17 @@ void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
             if (numberOfPointsDraggedOverThreshold >= 1) {
                 if (requiredNumberOfPointsDraggedOverThreshold && avgDrag.lengthSquared() >= dragThresholdSquared && accumulatedMovementMagnitude < dragThreshold) {
                     // Drag
-                    if (grabPoints(chosenPoints))
+                    if (grabPoints(event, chosenPoints))
                         setActive(true);
                 } else if (distanceToCentroidDelta > dragThreshold) {    // all points should in accumulation have been moved beyond threshold (?)
                     // Scale
-                    if (grabPoints(chosenPoints))
+                    if (grabPoints(event, chosenPoints))
                         setActive(true);
                 } else if (distanceToCentroidDelta < dragThreshold && (centroidMovementDelta < dragThreshold)) {
                     // Rotate
                     // Since it wasn't a scale and if we exceeded the dragthreshold, and the
                     // centroid didn't moved much, the points must have been moved around the centroid.
-                    if (grabPoints(chosenPoints))
+                    if (grabPoints(event, chosenPoints))
                         setActive(true);
                 }
             }
@@ -445,12 +402,12 @@ void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
         // avoid mapping the minima and maxima, as they might have unmappable values
         // such as -inf/+inf. Because of this we perform the bounding to min/max in local coords.
         // 1. scale
-        dist = averageTouchPointDistance(m_centroid.scenePosition());
+        dist = averageTouchPointDistance(centroid().scenePosition());
         m_activeScale = dist / m_startDistance;
         m_activeScale = qBound(m_minimumScale/m_startScale, m_activeScale, m_maximumScale/m_startScale);
 
         // 2. rotate
-        QVector<PointData> newAngles = angles(m_centroid.scenePosition());
+        QVector<PointData> newAngles = angles(centroid().scenePosition());
         const qreal angleDelta = averageAngleDelta(m_startAngles, newAngles);
         m_activeRotation += angleDelta;
         m_startAngles = std::move(newAngles);
@@ -465,25 +422,15 @@ void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
     m_accumulatedScale = m_startScale * m_activeScale;
 
     if (target() && target()->parentItem()) {
-        const QPointF centroidParentPos = target()->parentItem()->mapFromScene(m_centroid.scenePosition());
+        const QPointF centroidParentPos = target()->parentItem()->mapFromScene(centroid().scenePosition());
         // 3. Drag/translate
-        const QPointF centroidStartParentPos = target()->parentItem()->mapFromScene(m_centroid.sceneGrabPosition());
+        const QPointF centroidStartParentPos = target()->parentItem()->mapFromScene(centroid().sceneGrabPosition());
         m_activeTranslation = QVector2D(centroidParentPos - centroidStartParentPos);
         // apply rotation + scaling around the centroid - then apply translation.
-        QMatrix4x4 mat;
-
-        const QVector3D centroidParentVector(centroidParentPos);
-        mat.translate(centroidParentVector);
-        mat.rotate(float(m_activeRotation), 0, 0, 1);
-        mat.scale(float(m_activeScale));
-        mat.translate(-centroidParentVector);
-        mat.translate(QVector3D(m_activeTranslation));
-
-        mat = mat * m_startMatrix;
-
-        QPointF xformOriginPoint = target()->transformOriginPoint();
-        QPointF pos = mat * xformOriginPoint;
-        pos -= xformOriginPoint;
+        QPointF pos = QQuickItemPrivate::get(target())->adjustedPosForTransform(centroidParentPos,
+                                                                                m_startPos, m_activeTranslation,
+                                                                                m_startScale, m_activeScale,
+                                                                                m_startRotation, m_activeRotation);
 
         if (xAxis()->enabled())
             pos.setX(qBound(xAxis()->minimum(), pos.x(), xAxis()->maximum()));
@@ -498,12 +445,12 @@ void QQuickPinchHandler::handlePointerEventImpl(QQuickPointerEvent *event)
         target()->setRotation(rotation);
         target()->setScale(m_accumulatedScale);
     } else {
-        m_activeTranslation = QVector2D(m_centroid.scenePosition() - m_centroid.scenePressPosition());
+        m_activeTranslation = QVector2D(centroid().scenePosition() - centroid().scenePressPosition());
     }
 
-    qCDebug(lcPinchHandler) << "centroid" << m_centroid.scenePressPosition() << "->"  << m_centroid.scenePosition()
+    qCDebug(lcPinchHandler) << "centroid" << centroid().scenePressPosition() << "->"  << centroid().scenePosition()
                             << ", distance" << m_startDistance << "->" << dist
-                            << ", startScale" << m_startScale << "->" << m_accumulatedScale
+                            << ", scale" << m_startScale << "->" << m_accumulatedScale
                             << ", activeRotation" << m_activeRotation
                             << ", rotation" << rotation
                             << " from " << event->device()->type();
